#!/usr/bin/env perl

use XyceRegression::Tools;

$Tools = XyceRegression::Tools->new();
#$Tools->setDebug(1);
#$Tools->setVerbose(1);

# The input arguments to this script are: 
# $1 = location of Xyce binary
# $2 = location of xyce_verify.pl script
# $3 = location of compare script 
# $4 = location of circuit file to test
# $5 = location of gold standard prn file

# If Xyce does not produce a prn file, then we return exit code 10.
# If Xyce succeeds, but the test fails, then we return exit code 2.
# If the shell script fails for some reason, then we return exit code 1.
# Otherwise we return the exit code of compare or xyce_verify.pl

# Since the shell script runs Xyce and the comparison program, it is
# responsible for capturing any error output from Xyce and the STDOUT & STDERR
# from the comparison program.  The outside script, run_xyce_regression,
# expects the STDERR output from Xyce to go into $CIRFILE.err, the STDOUT
# output from comparison to go into $CIRFILE.prn.out and the STDERR output from
# comparison to go into $CIRFILE.prn.err.  

use Getopt::Long;

&GetOptions( "verbose!" => \$verbose );
$XYCE=$ARGV[0];
$XYCE_VERIFY=$ARGV[1];
#$XYCE_COMPARE=$ARGV[2];
$CIRFILE=$ARGV[3];
$GOLDPRN=$ARGV[4];

# these are the tolerances used in comparing the measures calculated by this 
# script to those calculated by Xyce.  Zerotol is needed because some measured
# values should be zero (and reltol won't make sense in that case).

# values measured here on the order of 50,000 so abstol needs to be big.
my $absTol = 1.0;
my $relTol = 0.02;
my $zeroTol = 1.0e-5;

# this is the tolerance used to tell if we are near enough to a time point 
# to start/stop a measure 
my $timeAbsTol = 1.0e-4;  

if (defined($verbose)) { $Tools->setVerbose(1); }

sub verbosePrint { $Tools->verbosePrint(@_); }

$retval = -1;
$retval=$Tools->wrapXyce($XYCE,$CIRFILE);
if ($retval != 0) { print "Exit code = $retval\n"; exit $retval; }
if (not -s "$CIRFILE.prn" ) { print "Exit code = 14\n"; exit 14; }

#
# Did we make a measure file
#
if (not -s "$CIRFILE.mt0" ) { print "Exit code = 17\n"; exit 17; }

#
# Now we need to count the number of .MEASURE statemets in the netlist
# and make sure we have some output for each one in the output file
#
open(NETLIST, "$CIRFILE");
$numMeasures=0;
while( $line=<NETLIST> )
{
  if( $line =~ /^\.measure/i)
  {  
    $numMeasures++;
  }
}
close(NETLIST);

#
# Now look for the measure output file and count how many measurements are in it
#
open(RESULTS, "$CIRFILE.mt0");
$numMeasuresFound=0;
my @measureNames;
my @measureVals;
while( $line=<RESULTS> )
{
  if( $line =~ /=/)
  {  
    # parse out NAME = Value 
    chop $line;
    ($name,$sep,$value) = split(/ /, $line);
    print "name = $name value=$value\n";
    push @measureNames, $name;
    push @measureVals, $value;
    $numMeasuresFound++;
  }
}
close(RESULTS);

if( $numMeasures != $numMeasuresFound )
{
  verbosePrint "Numer of requested measures did not match number of printed measures.\n";
  print "Exit code = 2\n"; exit 2; 
}

# Now look at Xyce's output file and read in Xyce's data
open(XYCE_RESULTS, "$CIRFILE.prn");
my @calculatedMeasureVals;
my @headerVarNames;
my @dataFromXyce;
my $numPts = 0;
while( $line=<XYCE_RESULTS> )
{
  if( $line =~ /TIME/)
  {  
    # parse out Header names 
    chop $line;
    @headerVarNames = split(/\s+/, $line);
  }
  elsif($line =~ /End of/)
  {
    # end of Xyce simulation line 
    # do nothing in this case
  }
  else
  {
    # save var values 
    chop $line;
    # Remove leading spaces on line, otherwise the spaces become element 0 of "@lineOfDataFromXyce" instead of the first column of data.
    $line =~ s/^\s*//;
    @lineOfDataFromXyce = (split(/\s+/, $line));
    for ($varNum=0; $varNum < @lineOfDataFromXyce; $varNum++)
    {
      #print "setting dataFromXyce[$numPts][$varNum] = $lineOfDataFromXyce[$varNum] \n";
      $dataFromXyce[$numPts][$varNum] = $lineOfDataFromXyce[$varNum];
    }
    $numPts++;
  }
}
close(XYCE_RESULTS);


# now calculate the values for these measures.  This is the part of 
# the test script that will be different for each one. 
# 
my @calMeasure;
$calMeasure[0]=0.0;
$calMeasure[1]=0.0;

# need to calculate slopes at specific times.
# keep track of (x1,y1) (x2,y2) data.
my @y1Vals;
my @x1Vals;
my @y2Vals;
my @x2Vals;

#
# I should be able to parse these from the netlist file on then
# individual measure lines "from x1 to x2"
$x1Vals[0] = 1.0e-3;
$x1Vals[1] = 1.4e-3;
$x2Vals[0] = 1.2e-3;
$x2Vals[1] = 1.6e-3;

my @point1Found;
my @point2Found;

$point1Found[0] = 0;
$point1Found[1] = 0;
$point2Found[0] = 0;
$point2Found[1] = 0;

for $i (1 .. $#dataFromXyce )
{  
  # find the time point Xyce output which 
  # is closest to the desired $x1Vals or $x2Vals
  for $j (0 .. 1)
  {
    if( $point1Found[$j] == 0 )
    {
      if( ($dataFromXyce[$i-1][0] <= $x1Vals[$j]) && ($x1Vals[$j] <= $dataFromXyce[$i][0]) )
      {
        $point1Found[$j] = 1;
        # pick next point 
        $x1Vals[$j] = $dataFromXyce[$i][0];
        $y1Vals[$j] = $dataFromXyce[$i][$j + 1];
        
        # pick the closest point
        #if( abs( $dataFromXyce[$i-1][0] - $x1Vals[$j] ) < abs($x1Vals[$j] - $dataFromXyce[$i][0]) )
        #{
        #  $x1Vals[$j] = $dataFromXyce[$i-1][0];
        #  $y1Vals[$j] = $dataFromXyce[$i-1][$j + 1];
        #}
        #else
        #{
        #  $x1Vals[$j] = $dataFromXyce[$i][0];
        #  $y1Vals[$j] = $dataFromXyce[$i][$j + 1];
        #}
      }
    }
    if( $point2Found[$j] == 0 )
    {
      if( ($dataFromXyce[$i-1][0] <= $x2Vals[$j]) && ($x2Vals[$j] <= $dataFromXyce[$i][0]) )
      {
        $point2Found[$j] = 1;
        # pick next point 
        $x2Vals[$j] = $dataFromXyce[$i-1][0];
        $y2Vals[$j] = $dataFromXyce[$i-1][$j + 1];
        
        # pick the closest point
        #if( abs( $dataFromXyce[$i-1][0] - $x2Vals[$j] ) < abs($x2Vals[$j] - $dataFromXyce[$i][0]) )
        #{
        #  $x2Vals[$j] = $dataFromXyce[$i-1][0];
        #  $y2Vals[$j] = $dataFromXyce[$i-1][$j + 1];
        #}
        #else
        #{
        #  $x2Vals[$j] = $dataFromXyce[$i][0];
        #  $y2Vals[$j] = $dataFromXyce[$i][$j + 1];
        #}
      }
    }
  }
}

my @absMeasureError;
my @relMeasureError;

# calculate the derivative
for $i (0 .. 1)
{
  $calMeasure[$i]=($y2Vals[$i] - $y1Vals[$i]) / ($x2Vals[$i] - $x1Vals[$i]);
  print "calculated measure = $calMeasure[$i]\n";
  $absMeasureError[$i] = abs( $measureVals[$i] - $calMeasure[$i] );
  if ( $calMeasure[$i] != 0 )
  {
    $relMeasureError[$i] = abs( $absMeasureError[$i] / $calMeasure[$i]);
  }
}

$retval = 0;
foreach $i (0 .. 1)
{
  print "For measure $measureNames[$i] calculated value = $calMeasure[$i]: abserror = $absMeasureError[$i] relerror=$relMeasureError[$i]\n";
  if( ($absMeasureError[$i] > $absTol) || (($measureVals[$i] > $zeroTol) && ($relMeasureError[$i] > $relTol) ))
  {
    print "Out of tolerance with abstol = $absTol and reltol = $relTol \n";
    $retval = 2;
  }
}


if ( $retval != 0 )
{
  verbosePrint "test Failed!\n";
  print "Exit code = $retval\n"; 
  exit $retval;
}
  
verbosePrint "test Passed!\n";
print "Exit code = $retval\n"; 
exit $retval;
